#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ncurses.h>
#include <menu.h>
#include <unistd.h>

#ifdef __FreeBSD__
#include <machine/soundcard.h>
#else
#include <sys/soundcard.h>
#endif

#include <stdlib.h>

#include "engine.h"
#include "config.h"
#include "interface.h"

#define ARLMTIME 1
#define VOLHASH 15
#define BALHASH 56
#define VOLARROW 1
#define ESC '\x1b'
#define XBACKSPACE '\x7f'

short int initInterface(void);
short int moveCursorAt(int position, int maxPosition);
void decreaseVolume(int device);
void increaseVolume(int device);
void leftBal(int device);
void rightBal(int device);
int setState(int device, char state);
void setTimer(struct sigaction *act, void(*func)(int signo));
void updateScreen(int signo);
void updateChannel(int device);
void saveSetting(void);
void loadSetting(void);
void deleteSetting(void);
void help(void);
void printError(WINDOW *win, const char *msg);

int cursmain() {
	short int currDevice=0;
	short int currRec;
	int ch;
	short int maxDevice;
	struct sigaction *act=(struct sigaction *)malloc(sizeof(struct sigaction));

	initMixer();
	initMask();
	getChannels();
	initscr();
	cbreak();
	noecho();
	keypad(stdscr, TRUE);
	curs_set(0);
	maxDevice=initInterface();
	mvaddch(1, VOLARROW, '>');
	refresh();
	setTimer(act, updateScreen);
	free(act);

	for(currRec=0; currRec <= maxDevice; currRec++) {
		if (devs[currRec].state == REC) break;
	}

	while((ch=getch()) != 'q') {
		switch(ch) {
			case KEY_UP:
				currDevice=(moveCursorAt(currDevice, maxDevice))-1;
				break;
			case KEY_DOWN:
				currDevice=(moveCursorAt(currDevice+2, maxDevice))-1;
				break;
			case KEY_LEFT:
				decreaseVolume(currDevice);
				break;
			case KEY_RIGHT:
				increaseVolume(currDevice);
				break;
			case 'a':
				leftBal(currDevice);
				break;
			case 'd':
				rightBal(currDevice);
				break;
			case 'm':
				if (devs[currDevice].state == MUTE) {
					if(!isRecording(devs[currDevice].chan))
						setState(currDevice, REC);
					else
						setState(currDevice, PLAY);
				}
				else
					setState(currDevice, MUTE);

				break;
			case 'r':
				if(!setState(currDevice, REC)) {
					if (currRec != currDevice) {
						setState(currRec, PLAY);
						mvaddch(currRec+1, VOLARROW, ' ');
						refresh();
					}
				currRec=currDevice;
				}
				break;
			case 'l':
				loadSetting();
				for(currRec=0; currRec <= maxDevice; currRec++) {
					if (devs[currRec].state == REC) break;
				}
				break;
			case 's':
				saveSetting();
				break;
			case 'e':
				deleteSetting();
				break;
			case 'h':
				help();
				break;
		}
	}

	endwin();
	exit(0);
}

short int initInterface()
{
	int counter;

	mvprintw(0,20, (char *)"---------------Ermixer-v0.8-------------");

	for(counter=0; devs[counter].chan != -1; counter++) {
		updateChannel(counter);
	}

	mvprintw(counter+1, 30, (char *)"Q to quit, H for help");
	refresh();

	return counter;
}

short int moveCursorAt(int position, int maxPosition)
{
	int newPosition;

	mvaddch(position-1, VOLARROW, ' ');
	mvaddch(position+1, VOLARROW, ' ');

	if (position > maxPosition) {
		newPosition=1;
	}
	else if (position < 1) {
		newPosition=maxPosition;
	}
	else newPosition=position;

	mvaddch(newPosition, VOLARROW, '>');
	refresh();

	return newPosition;
}

void increaseVolume(int device)
{
	if(devs[device].right < 100)
		devs[device].right++;
	if(devs[device].left < 100)
		devs[device].left++;
	if (devs[device].state != MUTE)
		writeToDevice(devs[device].chan, devs[device].left, devs[device].right);

	updateChannel(device);
}

void decreaseVolume(int device)
{
	if(devs[device].right > 0)
		devs[device].right--;
	if(devs[device].left > 0)
		devs[device].left--;
	if (devs[device].state != MUTE)
		writeToDevice(devs[device].chan, devs[device].left, devs[device].right);

	updateChannel(device);
}

void leftBal(int device)
{
	if(devs[device].type == STEREO) {
		if (devs[device].left < 100)
			devs[device].left+=2;

		if (((devs[device].right - devs[device].left) <= 0) && (devs[device].right > 0))
			devs[device].right-=2;

		writeToDevice(devs[device].chan, devs[device].left, devs[device].right);
		updateChannel(device);
	}
}

void rightBal(int device)
{
	if(devs[device].type == STEREO) {
		if (((devs[device].right - devs[device].left) >= 0) && (devs[device].left > 0))
			devs[device].left-=2;

		if (devs[device].right < 100)
			devs[device].right+=2;

		writeToDevice(devs[device].chan, devs[device].left, devs[device].right);
		updateChannel(device);
	}
}

int setState(int device, char state)
{
	switch(state) {
	case REC:
	          if(!isRecordable(devs[device].chan)) {
	          	modifySrc(devs[device].chan, state);
		  	writeToDevice(devs[device].chan, devs[device].left, devs[device].right);
		  }
		  else
		  	return ERROR;
		  break;
	case PLAY: writeToDevice(devs[device].chan, devs[device].left, devs[device].right); break;
	case MUTE: writeToDevice(devs[device].chan, 0, 0); break;
	}
	devs[device].state=state;
	updateChannel(device);
	return SUCCESS;
}

void updateChannel(int device)
{
	int a;
	int j;
	char c=0;
	const char *labels[] = SOUND_DEVICE_LABELS;

	switch (devs[device].state) {
		case REC: c='R'; break;
		case PLAY: c='P'; break;
		case PREC: c='P'; break;
		case MUTE: c='M'; break;
	}

	if(devs[device].type==STEREO)
		mvprintw(device+1, 2, (char *)" %-8s: -                      + |%3d|     Bal: A-            D+ |%+3d| %c ", labels[devs[device].chan], (devs[device].right + devs[device].left)/2, (devs[device].right - devs[device].left)/2, c);
	else
		mvprintw(device+1, 2, (char *)" %-8s: -                      + |%3d|                                 %c ", labels[devs[device].chan], devs[device].left, c);
	a=VOLHASH;

	for(j=(devs[device].right + devs[device].left)/10; j > 0; j--) {
		mvaddch(device+1, a, '#');
		a++;
	}

	if(devs[device].type==STEREO) {
		a=BALHASH;

		for(j=-5; j < ((devs[device].right - devs[device].left)/20); j++) {
			mvaddch(device+1, a, '#');
			a++;
		}
	}
	refresh();
}

void saveSetting()
{
	int c;
	char string[30];
	int i=0;
	WINDOW *saveDialog=newwin(15, 40, 4, 20);
	WINDOW *saveField=derwin(saveDialog, 1, 30, 4, 5);

	alarm(0);
	memset(string, 0, (sizeof(char)*30));
	
	keypad(saveDialog, TRUE);
	keypad(saveField, TRUE);

	box(saveDialog, 0,0);

	mvwprintw(saveDialog,0,15,(char *)"Save dialog");
	mvwprintw(saveDialog,2,8,(char *)"Enter profile name here");
	mvwprintw(saveDialog,6,12,(char *)"Press F10 to abort");
	wrefresh(saveDialog);
	wattron(saveField, A_UNDERLINE);
	mvwprintw(saveField,0,0,(char *)"%-30s", " ");
	curs_set(1);
	wmove(saveField, 0, 0);
	wrefresh(saveField);

	while((c=wgetch(saveField)) != '\n') {
		switch(c) {
			case KEY_LEFT:
				if(i > 0) {
					i--;
					wmove(saveField, 0, i);
				}
				break;
			case KEY_RIGHT:
				if(i < 30) {
					i++;
					wmove(saveField, 0, i);
				}
				break;
			case KEY_DC:
				wdelch(saveField);
				mvwaddch(saveField, 0, 29, ' ');
				wmove(saveField, 0, i);
				for(c=i; string[c] != '\0'; c++)
					string[c] = string[c+1];
				break;
			case KEY_BACKSPACE: 
			case XBACKSPACE:
				if(i > 0) {
					int j;
					i--;
					wmove(saveField, 0, i);
					wdelch(saveField);
					mvwaddch(saveField, 0, 29, ' ');
					wmove(saveField, 0, i);
					for(j=i; string[j] != '\0'; j++)
						string[j] = string[j+1];
				}
				break;
			case KEY_F(10):
				curs_set(0);
				delwin(saveDialog);
				redrawwin(stdscr);
				alarm(1);
				return;
			default:
				
				if((c >= 32 && c <= 255) && i < 30) {
					waddch(saveField, c);
					string[i]=c;
					i++;
					wmove(saveField, 0, i);
				}
				break;
		}
		wrefresh(saveField);
	}
	curs_set(0);
	if(strlen(string) == 0) {
		printError(saveDialog, "No name given");
		delwin(saveDialog);
		saveSetting();
		return;
	}
	c=writeFile(string);
	switch (c) {
		case NOSPACE:
			printError(saveDialog, "All 10 slot are full");
		case ERROR:
			printError(saveDialog, "Invalid file name");
			saveSetting();
	}

	delwin(saveDialog);
	redrawwin(stdscr);
	alarm(ARLMTIME);
}

void loadSetting()
{
	int i, c;
	ITEM **itemList=malloc(sizeof(ITEM *)*11);
	MENU *loadMenu;
	WINDOW *loadDialog=newwin(15, 40, 4, 20);

	alarm(0);
	keypad(loadDialog, TRUE);


	box(loadDialog, 0,0);
	mvwprintw(loadDialog,0,15,(char *)"Load dialog");

	getProfiles();

	for(i=0; i < 10; i++)
		itemList[i] = new_item(ret[i], " ");

	itemList[i] = (ITEM *) NULL;
	loadMenu=new_menu((ITEM **)itemList);
	set_menu_win(loadMenu, loadDialog);
	set_menu_sub(loadMenu, derwin(loadDialog, 10, 38, 2, 2));
	mvwprintw(loadDialog,13,5,(char *)"Return to select, Q to cancel");
	set_menu_mark(loadMenu, " > ");
	post_menu(loadMenu);
	wrefresh(loadDialog);
	i=0;

	while((c=getch()) != '\n') {
		switch(c) {
			case KEY_DOWN:
				if (!menu_driver(loadMenu, REQ_DOWN_ITEM))
					i++;
				wrefresh(loadDialog);
				break;
			case KEY_UP:
				if (!menu_driver(loadMenu, REQ_UP_ITEM))
					i--;
				wrefresh(loadDialog);
				break;
			case 'q':
				free_menu(loadMenu);

				for(i=0; itemList[i] != NULL; i++)
					free_item(itemList[i]);


				delwin(loadDialog);
				redrawwin(stdscr);
				alarm(1);
				return;
			default:
				break;
		}
	}

	parseFile(ret[i]);
	initMask();

	free_menu(loadMenu);

	for(i=0; itemList[i] != NULL; i++)
		free_item(itemList[i]);


	delwin(loadDialog);
	initInterface();
	redrawwin(stdscr);
	alarm(ARLMTIME);
}

void deleteSetting()
{
	int i, c;
	ITEM **itemList=malloc(sizeof(ITEM *)*11);
	MENU *deleteMenu;
	WINDOW *deleteDialog=newwin(15, 40, 4, 20);

	alarm(0);
	keypad(deleteDialog, TRUE);

	box(deleteDialog, 0,0);
	mvwprintw(deleteDialog,0,15,(char *)"Erase dialog");

	getProfiles();

	for(i=0; i < 10; i++)
		itemList[i] = new_item(ret[i], " ");

	itemList[i] = (ITEM *) NULL;
	deleteMenu=new_menu((ITEM **)itemList);
	set_menu_win(deleteMenu, deleteDialog);
	set_menu_sub(deleteMenu, derwin(deleteDialog, 10, 38, 2, 2));
	mvwprintw(deleteDialog,13,5,(char *)"Return to select, Q to cancel");
	set_menu_mark(deleteMenu, " > ");
	post_menu(deleteMenu);
	wrefresh(deleteDialog);
	i=0;

	while((c=getch()) != '\n') {
		switch(c) {
			case KEY_DOWN:
				if (!menu_driver(deleteMenu, REQ_DOWN_ITEM))
					i++;
				wrefresh(deleteDialog);
				break;
			case KEY_UP:
				if (!menu_driver(deleteMenu, REQ_UP_ITEM))
					i--;
				wrefresh(deleteDialog);
				break;
			case 'q':
				free_menu(deleteMenu);

				for(i=0; itemList[i] != NULL; i++)
					free_item(itemList[i]);


				delwin(deleteDialog);
				redrawwin(stdscr);
				alarm(1);
				return;

			default:
				break;
		}
	}

	deleteFile(ret[i]);

	free_menu(deleteMenu);

	for(i=0; itemList[i] != NULL; i++)
		free_item(itemList[i]);


	delwin(deleteDialog);
	deleteSetting();
	alarm(ARLMTIME);
}

void help()
{
	WINDOW *whelp=newwin(0,0,0,0);

	alarm(0);
	box(whelp,0,0);
	notimeout(whelp, TRUE);
	mvwprintw(whelp,1,30, (char *)"Ermixer Help");
	mvwprintw(whelp,2,5,(char *)"Right arrow : Increase volume");
	mvwprintw(whelp,3,5,(char *)"Left  arrow : Decrease volume");
	mvwprintw(whelp,4,5,(char *)"Upper arrow : Scroll to previous channel");
	mvwprintw(whelp,5,5,(char *)"Down  arrow : Scroll to next channel");
	mvwprintw(whelp,6,5,(char *)"Key M       : Mute\\Reactive Channel");
	mvwprintw(whelp,7,5,(char *)"Key R       : Set a device as the current recording device");
	mvwprintw(whelp,8,5,(char *)"Key S       : Open the save profiles dialog");
	mvwprintw(whelp,9,5,(char *)"Key L       : Open the load profiles dialog");
	mvwprintw(whelp,10,5,(char *)"Key E       : Open the erase profiles dialog");
	mvwprintw(whelp,11,5,(char *)"Key A       : Shift balancement to the left");
	mvwprintw(whelp,12,5,(char *)"Key D       : Shift balancement to the right");
	mvwprintw(whelp,13,5,(char *)"Key Q       : Quit");
	mvwprintw(whelp,14,5,(char *)"Key H       : Show this help");
	wrefresh(whelp);
	wgetch(whelp);
	delwin(whelp);
	redrawwin(stdscr);
	alarm(ARLMTIME);
}

void printError(WINDOW *win, const char *msg)
{
	WINDOW *errorWin=derwin(win, 7, 30, 5, 5);
	box(errorWin, 0, 0);
	notimeout(errorWin, TRUE);
	mvwprintw(errorWin,0,12,(char *)"ERROR");
	mvwprintw(errorWin,3,2, (char *)msg);
	wrefresh(errorWin);
	wgetch(errorWin);
	delwin(errorWin);
	redrawwin(win);
}

void setTimer(struct sigaction *act, void(*func)(int signo))
{
	act->sa_handler=func;
	sigemptyset(&act->sa_mask);
	act->sa_flags=0;
	#ifdef SA_RESTART
	act->sa_flags|=SA_RESTART;
	#endif
	sigaction(SIGALRM, act, NULL);
	alarm(ARLMTIME);
}


void updateScreen(int signo)
{
	signo=0; /* we don't need it */
    initMask();
	getChannels();
	initInterface();
	alarm(ARLMTIME);
}









